#define _GNU_SOURCE
#include <unistd.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <dirent.h>
#include <string.h>
#include <limits.h>

static uid_t (*real_getuid)(void);
static uid_t (*real_geteuid)(void);
static gid_t (*real_getgid)(void);
static gid_t (*real_getegid)(void);
static int (*real_setuid)(uid_t);
static int (*real_seteuid)(uid_t);
static int (*real_setreuid)(uid_t, uid_t);
static int (*real_setgid)(gid_t);
static int (*real_setegid)(gid_t);
static int (*real_setregid)(gid_t, gid_t);
static int (*real___xstat)(int, const char *, struct stat *);
static int (*real___lxstat)(int, const char *, struct stat *);
static int (*real___fxstat)(int, int, struct stat *);
static int (*real___xstat64)(int, const char *, struct stat64 *);
static int (*real___lxstat64)(int, const char *, struct stat64 *);
static int (*real___fxstat64)(int, int, struct stat64 *);
static int (*real_chown)(const char *, uid_t, gid_t);
static int (*real_lchown)(const char *, uid_t, gid_t);
static int (*real_fchown)(int, uid_t, gid_t);
static int (*real_unlink)(const char *);
static int (*real_remove)(const char *);
static int (*real_rmdir)(const char *);

static int do_pretend = 0;
static char *dir;

static void print_filename (const struct stat *st, char *res)
{
	sprintf(res, "%s/%04llx-%ld", dir, st->st_dev, st->st_ino);
}

static void print_filename64 (const struct stat64 *st, char *res)
{
	sprintf(res, "%s/%04llx-%lld", dir, st->st_dev, st->st_ino);
}

static void get_pretended_uid_gid (struct stat *st)
{
	char path[PATH_MAX];
	char buf[32];
	int fd, i;
	int uid, gid;

	uid = gid = 0;
	print_filename(st, path);
	fd = open(path, O_RDONLY);
	if (fd >= 0) {
		i = read(fd, buf, 32);
		if (i > 0 && i < 32) {
			buf[i] = 0;
			if (sscanf(buf, "%d %d", &uid, &gid) != 2) {
				uid = gid = 0;
			}
		}
		close(fd);
	}
	st->st_uid = uid;
	st->st_gid = gid;
}

static void get_pretended_uid_gid64 (struct stat64 *st)
{
	char path[PATH_MAX];
	char buf[32];
	int fd, i;
	int uid, gid;

	uid = gid = 0;
	print_filename64(st, path);
	fd = open(path, O_RDONLY);
	if (fd >= 0) {
		i = read(fd, buf, 32);
		if (i > 0 && i < 32) {
			buf[i] = 0;
			if (sscanf(buf, "%d %d", &uid, &gid) != 2) {
				uid = gid = 0;
			}
		}
		close(fd);
	}
	st->st_uid = uid;
	st->st_gid = gid;
}

static void remove_pretended_uid_gid (const struct stat *st)
{
	char path[PATH_MAX];
	print_filename(st, path);
	unlink(path);
}

static void set_pretended_uid_gid (struct stat *st, uid_t uid, gid_t gid)
{
	char path[PATH_MAX];
	char buf[32];
	int fd;

	get_pretended_uid_gid(st);
	if (uid != -1) st->st_uid = uid;
	if (gid != -1) st->st_gid = gid;
	remove_pretended_uid_gid(st);
	if (st->st_uid != 0 || st->st_gid != 0) {
		print_filename(st, path);
		fd = open(path, O_WRONLY|O_CREAT|O_EXCL, 0666);
		if (fd >= 0) {
			sprintf(buf, "%d %d\n", st->st_uid, st->st_gid);
			write(fd, buf, strlen(buf));
			close(fd);
		}
	}
}

void pretendroot_init (void) __attribute((constructor));
void pretendroot_init (void)
{
	real_getuid     = dlsym(RTLD_NEXT, "getuid");
	real_geteuid    = dlsym(RTLD_NEXT, "geteuid");
	real_getgid     = dlsym(RTLD_NEXT, "getgid");
	real_getegid    = dlsym(RTLD_NEXT, "getegid");
	real_setuid     = dlsym(RTLD_NEXT, "setuid");
	real_seteuid    = dlsym(RTLD_NEXT, "seteuid");
	real_setreuid   = dlsym(RTLD_NEXT, "setreuid");
	real_setgid     = dlsym(RTLD_NEXT, "setgid");
	real_setegid    = dlsym(RTLD_NEXT, "setegid");
	real_setregid   = dlsym(RTLD_NEXT, "setregid");
	real___xstat    = dlsym(RTLD_NEXT, "__xstat");
	real___lxstat   = dlsym(RTLD_NEXT, "__lxstat");
	real___fxstat   = dlsym(RTLD_NEXT, "__fxstat");
	real___xstat64  = dlsym(RTLD_NEXT, "__xstat64");
	real___lxstat64 = dlsym(RTLD_NEXT, "__lxstat64");
	real___fxstat64 = dlsym(RTLD_NEXT, "__fxstat64");
	real_chown      = dlsym(RTLD_NEXT, "chown");
	real_lchown     = dlsym(RTLD_NEXT, "lchown");
	real_fchown     = dlsym(RTLD_NEXT, "fchown");
	real_unlink     = dlsym(RTLD_NEXT, "unlink");
	real_remove     = dlsym(RTLD_NEXT, "remove");
	real_rmdir      = dlsym(RTLD_NEXT, "rmdir");

	dir = getenv("PRETENDROOTDIR");
	do_pretend = (dir != NULL);

	if (do_pretend) {
		if (getenv("USER")     != NULL) setenv("USER",     "root", 1);
		if (getenv("USERNAME") != NULL) setenv("USERNAME", "root", 1);
		if (getenv("LOGNAME")  != NULL) setenv("LOGNAME",  "root", 1);
	}
}

uid_t getuid (void)
{
	if (!do_pretend) return real_getuid();
	return 0;
}

uid_t geteuid (void)
{
	if (!do_pretend) return real_geteuid();
	return 0;
}

gid_t getgid (void)
{
	if (!do_pretend) return real_getgid();
	return 0;
}

gid_t getegid (void)
{
	if (!do_pretend) return real_getegid();
	return 0;
}

int setuid (uid_t uid)
{
	if (!do_pretend) return real_setuid(uid);
	do_pretend = 0;
	real_setuid(uid);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int seteuid (uid_t euid)
{
	if (!do_pretend) return real_seteuid(euid);
	do_pretend = 0;
	real_seteuid(euid);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int setreuid (uid_t ruid, uid_t euid)
{
	if (!do_pretend) return real_setreuid(ruid, euid);
	do_pretend = 0;
	real_setreuid(ruid, euid);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int setgid (gid_t gid)
{
	if (!do_pretend) return real_setgid(gid);
	do_pretend = 0;
	real_setgid(gid);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int setegid (gid_t egid)
{
	if (!do_pretend) return real_setegid(egid);
	do_pretend = 0;
	real_setegid(egid);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int setregid (gid_t rgid, gid_t egid)
{
	if (!do_pretend) return real_setregid(rgid, egid);
	do_pretend = 0;
	real_setregid(rgid, egid);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int __xstat (int ver, const char *file, struct stat *st)
{
	if (!do_pretend) return real___xstat(ver, file, st);
	do_pretend = 0;
	if (real___xstat(ver, file, st) < 0) {
		do_pretend = 1;
		return -1;
	}
	get_pretended_uid_gid(st);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int __lxstat (int ver, const char *file, struct stat *st)
{
	if (!do_pretend) return real___lxstat(ver, file, st);
	do_pretend = 0;
	if (real___lxstat(ver, file, st) < 0) {
		do_pretend = 1;
		return -1;
	}
	get_pretended_uid_gid(st);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int __fxstat (int ver, int fd, struct stat *st)
{
	if (!do_pretend) return real___fxstat(ver, fd, st);
	do_pretend = 0;
	if (real___fxstat(ver, fd, st) < 0) {
		do_pretend = 1;
		return -1;
	}
	get_pretended_uid_gid(st);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int __xstat64 (int ver, const char *file, struct stat64 *st)
{
	if (!do_pretend) return real___xstat64(ver, file, st);
	do_pretend = 0;
	if (real___xstat64(ver, file, st) < 0) {
		do_pretend = 1;
		return -1;
	}
	get_pretended_uid_gid64(st);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int __lxstat64 (int ver, const char *file, struct stat64 *st)
{
	if (!do_pretend) return real___lxstat64(ver, file, st);
	do_pretend = 0;
	if (real___lxstat64(ver, file, st) < 0) {
		do_pretend = 1;
		return -1;
	}
	get_pretended_uid_gid64(st);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int __fxstat64 (int ver, int fd, struct stat64 *st)
{
	if (!do_pretend) return real___fxstat64(ver, fd, st);
	do_pretend = 0;
	if (real___fxstat64(ver, fd, st) < 0) {
		do_pretend = 1;
		return -1;
	}
	get_pretended_uid_gid64(st);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int chown (const char *file, uid_t uid, gid_t gid)
{
	struct stat st;
	if (!do_pretend) return real_chown(file, uid, gid);
	do_pretend = 0;
	if (stat(file, &st) < 0) {
		do_pretend = 1;
		return -1;
	}
	set_pretended_uid_gid(&st, uid, gid);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int lchown (const char *file, uid_t uid, gid_t gid)
{
	struct stat st;
	if (!do_pretend) return real_lchown(file, uid, gid);
	do_pretend = 0;
	if (lstat(file, &st) < 0) {
		do_pretend = 1;
		return -1;
	}
	set_pretended_uid_gid(&st, uid, gid);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int fchown (int fd, uid_t uid, gid_t gid)
{
	struct stat st;
	if (!do_pretend) return real_fchown(fd, uid, gid);
	do_pretend = 0;
	if (fstat(fd, &st) < 0) {
		do_pretend = 1;
		return -1;
	}
	set_pretended_uid_gid(&st, uid, gid);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int unlink (const char *file)
{
	struct stat st;
	if (!do_pretend) return real_unlink(file);
	do_pretend = 0;
	if (lstat(file, &st) < 0) {
		do_pretend = 1;
		return -1;
	}
	if (real_unlink(file) < 0) {
		do_pretend = 1;
		return -1;
	}
	if (st.st_nlink == 1) remove_pretended_uid_gid(&st);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int remove (const char *file)
{
	struct stat st;
	if (!do_pretend) return real_remove(file);
	do_pretend = 0;
	if (lstat(file, &st) < 0) {
		do_pretend = 1;
		return -1;
	}
	if (real_remove(file) < 0) {
		do_pretend = 1;
		return -1;
	}
	if (st.st_nlink == 1) remove_pretended_uid_gid(&st);
	do_pretend = 1;
	errno = 0;
	return 0;
}

int rmdir (const char *file)
{
	struct stat st;
	if (!do_pretend) return real_rmdir(file);
	do_pretend = 0;
	if (lstat(file, &st) < 0) {
		do_pretend = 1;
		return -1;
	}
	if (real_rmdir(file) < 0) {
		do_pretend = 1;
		return -1;
	}
	remove_pretended_uid_gid(&st);
	do_pretend = 1;
	errno = 0;
	return 0;
}

